/*
 * Access several filesystems.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <map>

#include <fsclient/fsdesc.h>
#include <fsclient/fsdesc_client.h>
#include "fspath.h"

/* Exposure of the resource. */

#include <fsserver/resource_server.h>

/* Implemented interfaces. */

#include <fsserver/user_filesystem_resource.h>



/* Mountpoint-to-server mapping. */

typedef std::map<const char *, l4_cap_idx_t> MountTable;
typedef std::map<const char *, l4_cap_idx_t>::iterator MountTableIterator;
typedef std::map<const char *, l4_cap_idx_t>::value_type MountTableEntry;



/* Virtual filesystem resource. */

class Vfs_resource : public UserFilesystemResource
{
private:
  MountTable _servers;

public:
  explicit Vfs_resource()
  : UserFilesystemResource()
  {
  }



  /* Obtain an object as a filesystem reference. */

  long getfs_object(fs_object_t *fsobj, l4_cap_idx_t *fs)
  {
    file_descriptor_t desc;
    long err;

    /* Obtain a suitable descriptor. */

    err = init_desc(&desc, fsobj);

    if (err)
      return err;

    /* Return a reference to any directly-accessed object. */

    char *path = fsdesc_get_name(&fsobj->buffer);

    if ((path != NULL) && !strlen(path))
    {
      *fs = desc.server;
      return L4_EOK;
    }

    /* Obtain a context. */

    err = fs_context(&desc);

    if (err)
      return err;

    /* Find the path within the selected filesystem. */

    strcpy(desc.cobj.addr, fsobj->buffer.addr);

    /* Propagate the message to the identified filesystem.
       The L4_MSGTAG_PROPAGATE flag is not supported by Fiasco.OC, so this
       object has to act as intermediary. */

    err = fs_context_getfs(&desc);

    if (err)
    {
      fsdesc_deallocate_context(&desc);
      return err;
    }

    /* Free the context and its buffer from this task. */

    fsdesc_deallocate_context(&desc);

    /* Export and eventually free the server reference from this task. */

    *fs = discard_cap(desc.ref);

    return L4_EOK;
  }

  /* Mount a filesystem at a particular location. */

  long mount_object(fs_object_t *fsobj, l4_cap_idx_t fs)
  {
    /* Obtain the object's path. */

    char *path = fsdesc_get_name(&fsobj->buffer);

    if (path == NULL)
      return -L4_EINVAL;

    /* Mount the supplied capability. */

    mount_at_path(fs, path);

    return L4_EOK;
  }

  /* Open a filesystem object. */

  long open_object(fs_object_t *fsobj, int flags, size_t *size, l4_cap_idx_t *file)
  {
    file_descriptor_t desc;
    long err;

    /* Obtain a suitable descriptor. */

    err = init_desc(&desc, fsobj);

    if (err)
      return err;

    /* Obtain a context. */

    err = fs_context(&desc);

    if (err)
      return err;

    /* Find the path within the selected filesystem. */

    strcpy(desc.cobj.addr, fsobj->buffer.addr);

    /* Propagate the message to the identified filesystem.
       The L4_MSGTAG_PROPAGATE flag is not supported by Fiasco.OC, so this
       object has to act as intermediary. */

    err = fs_context_open(&desc, flags);

    if (err)
    {
      fsdesc_deallocate_context(&desc);
      return err;
    }

    /* Free the context and its buffer from this task. */

    fsdesc_deallocate_context(&desc);

    /* Export and eventually free the server reference from this task. */

    *size = desc.obj.size;
    *file = discard_cap(desc.ref);

    return L4_EOK;
  }



  /* Non-exported methods. */

  /* Initialise a descriptor for accessing a mounted filesystem. */

  long init_desc(file_descriptor_t *desc, fs_object_t *fsobj)
  {
    /* Match the path to a mountpoint. */

    char *path = fsdesc_get_name(&fsobj->buffer);

    if (path == NULL)
      return -L4_EINVAL;

    MountTableIterator entry = find_mountpoint(path);

    if (entry == _servers.end())
      return -L4_ENOENT;

    /* Rewrite the path for the selected filesystem. */

    rewrite_path(path, entry->first);

    /* Use the filesystem as the target of any query. */

    desc->server = entry->second;

    return L4_EOK;
  }

  void mount_at_path(l4_cap_idx_t server, const char *path)
  {
    /* NOTE: Should terminate with path separator if absent.
       NOTE: Should also ensure zero termination. */

    _servers.insert(MountTableEntry(strdup(path), server));
  }

  MountTableIterator find_mountpoint(const char *path)
  {
    /* Search only until "later" paths are encountered. Such paths will either
       incorporate components that are themselves greater than those in the path
       or will be longer than the path. */

    MountTableIterator it, found = _servers.end(),
                           limit = _servers.upper_bound(path);
    ssize_t length, longest = 0;

    /* Find the largest matching path. */

    for (it = _servers.begin(); it != limit; it++)
    {
      length = match_path(path, it->first);

      if (length >= longest)
        found = it;
    }

    return found;
  }

  ssize_t match_path(const char *path, const char *mountpoint)
  {
    size_t length = strlen(mountpoint);
    int suffixed = (mountpoint[length - 1] == '/');

    /* The empty mountpoint always matches;
       suffixed mountpoints may match as prefixes;
       suffixed mountpoints may match unsuffixed paths;
       unsuffixed mountpoints must match exactly. */

    /* NOTE: Unsuffixed mountpoints might support subpaths by insisting on
             a path separator as the first subsequent character. */

    if (!length ||
        (suffixed && !strncmp(mountpoint, path, length)) ||
        (suffixed && (strlen(path) == length - 1) && !strncmp(mountpoint, path, length - 1)) ||
        !strcmp(mountpoint, path))
      return length;

    return -1;
  }
};



int main(int argc, char *argv[])
{
  FILE *fp;
  char buffer[256], *sep;
  l4_cap_idx_t fscap;

  if (argc < 2)
  {
    printf("Need a mount table.\n");
    return 1;
  }

  /* Initialise and register a new server object. */

  Vfs_resource resource;
  ResourceServer server(&resource);

  if (server.bind("export"))
  {
    printf("Could not bind thread.\n");
    return 1;
  }

  /* Mount filesystems. */

  fp = fopen(argv[1], "r");
  if (fp == NULL)
  {
    printf("Could not open %s.\n", argv[1]);
    return 1;
  }

  /* Obtain capabilities for mountpoints. */

  while (fgets(buffer, 256, fp) != NULL)
  {
    sep = strchr(buffer, (int) '\t');
    if (sep == NULL)
      continue;

    /* Terminate newline-suffixed strings. */

    if (sep[strlen(sep) - 1] == '\n')
      sep[strlen(sep) - 1] = '\0';

    /* Obtain the named capability. */

    fscap = l4re_env_get_cap(sep + 1);

    if (l4_is_invalid_cap(fscap))
      continue;

    /* Terminate the path and register the mountpoint. */

    *sep = '\0';
    resource.mount_at_path(fscap, buffer);
  }

  fclose(fp);

  /* Enter the IPC server loop. */

  printf("Starting vfs resource...\n");
  server.start();
  return 0;
}
